/* ----------------------------------------------------------------------------
 * gamepad_mouse version 0.1
 * emulates mouse using gamepad
 *
 * Copyright 2004 Takuo Yonezawa (takuo@aya.or.jp)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <linux/input.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "linux/uinput.h"
#include "main.h"

char *name = "gamepad_mouse";
char *version_information =
  "gamepad_mouse version 0.1\n"
  "\n"
  "Copyright (C) 2004 Takuo Yonezawa (takuo@aya.or.jp)\n"
  "\n"
  "This is free software; see the source for copying conditions.\n"
  "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n";
char *description = "Convert gamepad event from device to mouse motion.\n";
#define MOUSE_SCALE 10

typedef enum {
  WHEEL_POSITION_CENTER,
  WHEEL_POSITION_TOP,
  WHEEL_POSITION_TOPRIGHT,
  WHEEL_POSITION_RIGHT,
  WHEEL_POSITION_BOTTOMRIGHT,
  WHEEL_POSITION_BOTTOM,
  WHEEL_POSITION_BOTTOMLEFT,
  WHEEL_POSITION_LEFT,
  WHEEL_POSITION_TOPLEFT,
} WheelPosition;

typedef struct _GamepadMouseState GamepadMouseState;
struct _GamepadMouseState 
{
  WheelPosition wheel_position;
  int wheel_x, wheel_y;
  int vx, vy;
  int32_t width_x, width_y, width_z, width_rz;
  int32_t center_x, center_y, center_z, center_rz;
};

static GamepadMouseState state;
static struct itimerval itimerval;
static struct itimerval old_itimerval;

void alarm_handler(int signum)
{
  if (state.vx == 0 && state.vy == 0) {
    return;
  }

  if (state.vx!= 0) {
    int movement;
    
    if (state.vx * 2 - 1 == state.width_z && state.width_z % 2 == 1) {
      state.vx--;
    }
    
    movement = state.vx * MOUSE_SCALE * 2 / state.width_z;
    send_event(EV_REL, REL_X, movement);
  }

  if (state.vy!= 0) {
    int movement;
    
    if (state.vy * 2 - 1 == state.width_rz && state.width_rz % 2 == 1) {
      state.vy--;
    }
    
    movement = state.vy * MOUSE_SCALE * 2 / state.width_rz;
    send_event(EV_REL, REL_Y, movement);
  }
}

void *init_event_handler(int number_of_event_device_files, int *event_fds)
{
  int32_t absolute_feature[4];

  state.wheel_position = WHEEL_POSITION_CENTER;
  state.wheel_x = 0;
  state.wheel_y = 0;
  state.vx = 0;
  state.vy = 0;
  ioctl(event_fds[0], EVIOCGABS(ABS_X), absolute_feature);
  state.width_x = absolute_feature[2] - absolute_feature[1];
  state.center_x = (absolute_feature[2] + absolute_feature[1]) / 2;
  ioctl(event_fds[0], EVIOCGABS(ABS_Y), absolute_feature);
  state.width_y = absolute_feature[2] - absolute_feature[1];
  state.center_y = (absolute_feature[2] + absolute_feature[1]) / 2;
  ioctl(event_fds[0], EVIOCGABS(ABS_Z), absolute_feature);
  state.width_z = absolute_feature[2] - absolute_feature[1];
  state.center_z = (absolute_feature[2] + absolute_feature[1]) / 2;
  ioctl(event_fds[0], EVIOCGABS(ABS_RZ), absolute_feature);
  state.width_rz = absolute_feature[2] - absolute_feature[1];
  state.center_rz = (absolute_feature[2] + absolute_feature[1]) / 2;

  itimerval.it_value.tv_sec = 0;
  itimerval.it_value.tv_usec = 50000;
  itimerval.it_interval.tv_sec = 0;
  itimerval.it_interval.tv_usec = 50000;
  signal(SIGALRM, alarm_handler);
  setitimer(ITIMER_REAL, &itimerval, &old_itimerval);

  return (void *)&state;
}

void finalize_event_hander(void *user_data)
{
  setitimer(ITIMER_REAL, &old_itimerval, NULL);
  signal(SIGALRM, SIG_DFL);
}

void update_wheel(GamepadMouseState *state)
{
  int x = state->wheel_x;
  int y = state->wheel_y;
  int movement = 0;
  WheelPosition new_wheel_position;
  static int movement_table[][9] =
    {
      {0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 1, 2, 0, 0, 0, -2, -1},
      {0, -1, 0, 1, 2, 0, 0, 0, -2},
      {0, -2, -1, 0, 1, 2, 0, 0, 0},
      {0, 0, -2, -1, 0, 1, 2, 0, 0},
      {0, 0, 0, -2, -1, 0, 1, 2, 0},
      {0, 0, 0, 0, -2, -1, 0, 1, 2},
      {0, 2, 0, 0, 0, -2, -1, 0, 1},
      {0, 1, 2, 0, 0, 0, -2, -1, 0},
    };
  
  if (abs(x) < state->width_x / 2 * 4 / 5 &&
      abs(y) < state->width_y / 2 * 4 / 5) {
    new_wheel_position = WHEEL_POSITION_CENTER;
  } else if (abs(x) > 2 * abs(y)) {
    if (x > 0) {
      new_wheel_position = WHEEL_POSITION_RIGHT;
    } else {
      new_wheel_position = WHEEL_POSITION_LEFT;
    }
  } else if (2 * abs(x) < abs(y)) {
    if (y > 0) {
      new_wheel_position = WHEEL_POSITION_BOTTOM;
    } else {
      new_wheel_position = WHEEL_POSITION_TOP;
    }
  } else {
    if (y > 0) {
      if (x > 0) {
        new_wheel_position = WHEEL_POSITION_BOTTOMRIGHT;
      } else {
        new_wheel_position = WHEEL_POSITION_BOTTOMLEFT;
      }
    } else {
      if (x > 0) {
        new_wheel_position = WHEEL_POSITION_TOPRIGHT;
      } else {
        new_wheel_position = WHEEL_POSITION_TOPLEFT;
      }
    }
  }
  movement = movement_table[state->wheel_position][new_wheel_position];
  state->wheel_position = new_wheel_position;
  if (movement != 0) {
    report_scroll(movement);
  }
}

void handle_event(int fd, struct input_event event, void *user_data)
{
  GamepadMouseState *state = (GamepadMouseState *)user_data;
 
  switch(event.type) {
    case EV_ABS:
      switch(event.code) {
        case ABS_X:
          state->wheel_x = (int)event.value - state->center_x;
          update_wheel(state);
          break;
        case ABS_Y:
          state->wheel_y = (int)event.value - state->center_y;
          update_wheel(state);
          break;
        case ABS_Z:
          state->vx = (int)event.value - state->center_z;
          break;
        case ABS_RZ:
          state->vy = (int)event.value - state->center_rz;
          break;
        default:
          break;
      }
      break;
    case EV_KEY:
      switch(event.code) {
        case BTN_PINKIE:
          send_event(EV_KEY, BTN_LEFT, event.value);
          break;
        case BTN_BASE2:
          send_event(EV_KEY, BTN_RIGHT, event.value);
          break;
        case BTN_THUMB:
          send_event(EV_KEY, KEY_HOME, event.value);
          break;
        case BTN_TOP:
          send_event(EV_KEY, KEY_END, event.value);
          break;
        case BTN_TRIGGER:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_W, 1);
            send_event(EV_KEY, KEY_W, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        case BTN_THUMB2:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_T, 1);
            send_event(EV_KEY, KEY_T, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        case BTN_TOP2:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_PAGEUP, 1);
            send_event(EV_KEY, KEY_PAGEUP, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        case BTN_BASE:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_PAGEDOWN, 1);
            send_event(EV_KEY, KEY_PAGEDOWN, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        default:
          break;
      }
      break;
    default:
      break;
  }
}

void set_event_bits(int ui_fd)
{  
	ioctl(ui_fd, UI_SET_EVBIT, EV_REL);
  ioctl(ui_fd, UI_SET_RELBIT, REL_WHEEL);
  ioctl(ui_fd, UI_SET_RELBIT, REL_X);
  ioctl(ui_fd, UI_SET_RELBIT, REL_Y);

  ioctl(ui_fd, UI_SET_EVBIT, EV_KEY);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_LEFTCTRL);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_HOME);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_END);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_PAGEUP);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_PAGEDOWN);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_W);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_T);
  ioctl(ui_fd, UI_SET_KEYBIT, BTN_LEFT);
  ioctl(ui_fd, UI_SET_KEYBIT, BTN_RIGHT);
}
