/* ----------------------------------------------------------------------------
 * gamepad_mouse2 version 0.1
 * emulates mouse using gamepad
 *
 * Copyright 2004 Takuo Yonezawa (takuo@aya.or.jp)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <linux/input.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "linux/uinput.h"
#include "main.h"

char *name = "gamepad_mouse";
char *version_information =
  "gamepad_mouse2 version 0.1\n"
  "\n"
  "Copyright (C) 2004 Takuo Yonezawa (takuo@aya.or.jp)\n"
  "\n"
  "This is free software; see the source for copying conditions.\n"
  "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n";
char *description = "Convert gamepad event from device to mouse motion.\n";
#define MOUSE_SCALE 5

typedef enum {
  STICK_POSITION_CENTER,
  STICK_POSITION_TOP,
  STICK_POSITION_TOPRIGHT,
  STICK_POSITION_RIGHT,
  STICK_POSITION_BOTTOMRIGHT,
  STICK_POSITION_BOTTOM,
  STICK_POSITION_BOTTOMLEFT,
  STICK_POSITION_LEFT,
  STICK_POSITION_TOPLEFT,
} StickPosition;

typedef struct _Stick Stick;
struct _Stick
{
  StickPosition position;
  int x, y;
  int32_t width, height;
  int32_t center_x, center_y;
};

typedef struct _GamepadMouseState GamepadMouseState;
struct _GamepadMouseState 
{
  Stick wheel;
  Stick cursor;
};

static GamepadMouseState state;

void *init_event_handler(int number_of_event_device_files, int *event_fds)
{
  int32_t absolute_feature[4];

  state.wheel.position = STICK_POSITION_CENTER;
  ioctl(event_fds[0], EVIOCGABS(ABS_X), absolute_feature);
  state.wheel.width = absolute_feature[2] - absolute_feature[1];
  state.wheel.center_x = (absolute_feature[2] + absolute_feature[1]) / 2;
  state.wheel.x = state.wheel.center_x;
  ioctl(event_fds[0], EVIOCGABS(ABS_Y), absolute_feature);
  state.wheel.height = absolute_feature[2] - absolute_feature[1];
  state.wheel.center_y = (absolute_feature[2] + absolute_feature[1]) / 2;
  state.wheel.y = state.wheel.center_y;

  state.cursor.position = STICK_POSITION_CENTER;
  ioctl(event_fds[0], EVIOCGABS(ABS_Z), absolute_feature);
  state.cursor.width = absolute_feature[2] - absolute_feature[1];
  state.cursor.center_x = (absolute_feature[2] + absolute_feature[1]) / 2;
  state.cursor.x = state.cursor.center_x;
  ioctl(event_fds[0], EVIOCGABS(ABS_RZ), absolute_feature);
  state.cursor.height = absolute_feature[2] - absolute_feature[1];
  state.cursor.center_y = (absolute_feature[2] + absolute_feature[1]) / 2;
  state.cursor.y = state.cursor.center_y;

  return (void *)&state;
}

void finalize_event_hander(void *user_data)
{
}

StickPosition calc_new_position(Stick *stick)
{
  int x = stick->x;
  int y = stick->y;
  int32_t width = stick->width;
  int32_t height = stick->height;

  if (abs(x) < width / 2 * 4 / 5 &&
      abs(y) < height / 2 * 4 / 5) {
    return STICK_POSITION_CENTER;
  } else if (abs(x) > 2 * abs(y)) {
    if (x > 0) {
      return STICK_POSITION_RIGHT;
    } else {
      return STICK_POSITION_LEFT;
    }
  } else if (2 * abs(x) < abs(y)) {
    if (y > 0) {
      return STICK_POSITION_BOTTOM;
    } else {
      return STICK_POSITION_TOP;
    }
  } else {
    if (y > 0) {
      if (x > 0) {
        return STICK_POSITION_BOTTOMRIGHT;
      } else {
        return STICK_POSITION_BOTTOMLEFT;
      }
    } else {
      if (x > 0) {
        return STICK_POSITION_TOPRIGHT;
      } else {
        return STICK_POSITION_TOPLEFT;
      }
    }
  }
}

static int cursor_movement_table_x[] = {0, 0, 1, 1, 1, 0, -1, -1, -1};
static int cursor_movement_table_y[] = {0, -1, -1, 0, 1, 1, 1, 0, -1};

void update_cursor(GamepadMouseState *state)
{
  StickPosition new_cursor_position;
  int movement_x, movement_y;
  
  new_cursor_position = calc_new_position(&(state->cursor));
  if (state->cursor.position == STICK_POSITION_CENTER) {
    movement_x = cursor_movement_table_x[new_cursor_position];
    movement_y = cursor_movement_table_y[new_cursor_position];
    if (movement_x != 0){
      send_event(EV_REL, REL_X, movement_x);
    }
    if (movement_y != 0){
      send_event(EV_REL, REL_Y, movement_y);
    }
  }
  state->cursor.position = new_cursor_position;
}

void update_wheel(GamepadMouseState *state)
{
  int movement = 0;
  StickPosition new_wheel_position;
  static int movement_table[][9] =
    {
      {0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 1, 2, 0, 0, 0, -2, -1},
      {0, -1, 0, 1, 2, 0, 0, 0, -2},
      {0, -2, -1, 0, 1, 2, 0, 0, 0},
      {0, 0, -2, -1, 0, 1, 2, 0, 0},
      {0, 0, 0, -2, -1, 0, 1, 2, 0},
      {0, 0, 0, 0, -2, -1, 0, 1, 2},
      {0, 2, 0, 0, 0, -2, -1, 0, 1},
      {0, 1, 2, 0, 0, 0, -2, -1, 0},
    };

  new_wheel_position = calc_new_position(&(state->wheel));
  movement = movement_table[state->wheel.position][new_wheel_position];
  state->wheel.position = new_wheel_position;
  if (movement != 0) {
    if (state->cursor.position == STICK_POSITION_CENTER) {
      report_scroll(movement);
    } else  {
      int movement_x, movement_y;

      movement_x = cursor_movement_table_x[state->cursor.position];
      movement_y = cursor_movement_table_y[state->cursor.position];
      if (movement_x != 0) {
        send_event(EV_REL, REL_X, movement_x * MOUSE_SCALE * movement);
      }
      if (movement_y != 0) {
        send_event(EV_REL, REL_Y, movement_y * MOUSE_SCALE * movement);
      }
    }
  }
}

void handle_event(int fd, struct input_event event, void *user_data)
{
  GamepadMouseState *state = (GamepadMouseState *)user_data;
 
  switch(event.type) {
    case EV_ABS:
      switch(event.code) {
        case ABS_X:
          state->wheel.x = (int)event.value - state->wheel.center_x;
          update_wheel(state);
          break;
        case ABS_Y:
          state->wheel.y = (int)event.value - state->wheel.center_y;
          update_wheel(state);
          break;
        case ABS_Z:
          state->cursor.x = (int)event.value - state->cursor.center_x;
          update_cursor(state);
          break;
        case ABS_RZ:
          state->cursor.y = (int)event.value - state->cursor.center_y;
          update_cursor(state);
          break;
        default:
          break;
      }
      break;
    case EV_KEY:
      switch(event.code) {
        case BTN_PINKIE:
          send_event(EV_KEY, BTN_LEFT, event.value);
          break;
        case BTN_BASE2:
          send_event(EV_KEY, BTN_RIGHT, event.value);
          break;
        case BTN_THUMB:
          send_event(EV_KEY, KEY_HOME, event.value);
          break;
        case BTN_TOP:
          send_event(EV_KEY, KEY_END, event.value);
          break;
        case BTN_TRIGGER:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_W, 1);
            send_event(EV_KEY, KEY_W, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        case BTN_THUMB2:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_T, 1);
            send_event(EV_KEY, KEY_T, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        case BTN_TOP2:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_PAGEUP, 1);
            send_event(EV_KEY, KEY_PAGEUP, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        case BTN_BASE:
          if (event.value == 0) {
            send_event(EV_KEY, KEY_LEFTCTRL, 1);
            send_event(EV_KEY, KEY_PAGEDOWN, 1);
            send_event(EV_KEY, KEY_PAGEDOWN, 0);
            send_event(EV_KEY, KEY_LEFTCTRL, 0);
          }
          break;
        default:
          break;
      }
      break;
    default:
      break;
  }
}

void set_event_bits(int ui_fd)
{  
	ioctl(ui_fd, UI_SET_EVBIT, EV_REL);
  ioctl(ui_fd, UI_SET_RELBIT, REL_WHEEL);
  ioctl(ui_fd, UI_SET_RELBIT, REL_X);
  ioctl(ui_fd, UI_SET_RELBIT, REL_Y);

  ioctl(ui_fd, UI_SET_EVBIT, EV_KEY);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_LEFTCTRL);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_HOME);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_END);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_PAGEUP);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_PAGEDOWN);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_W);
  ioctl(ui_fd, UI_SET_KEYBIT, KEY_T);
  ioctl(ui_fd, UI_SET_KEYBIT, BTN_LEFT);
  ioctl(ui_fd, UI_SET_KEYBIT, BTN_RIGHT);
}
